/*---------------------------------------------------------------------------*\

    Copyright (C) 2000 AVM GmbH. All rights reserved.

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY, without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE, and WITHOUT
    ANY LIABILITY FOR ANY DAMAGES arising out of or in connection
    with the use or performance of this software. See the
    GNU General Public License for further details.

\*---------------------------------------------------------------------------*/

#include "CapiChan.h"
#include "ConvCapi.h"
#include "atomar.h"
#include "dbgCAPI.h"

/*---------------------------------------------------------------------------*\
\*---------------------------------------------------------------------------*/

#define TIMER_FOR_RINGING                    1
#define TIMER_FOR_BCHANNEL                   2
#define TIMER_FOR_DISCONNECT                 3
#define TIMER_BCHANNEL_WAITTIME             90 // sec
#define TIMER_DISCONNECT_WAITTIME           10 // sec

static void StateChangeTimer_CallBack (void *pCapiChannel, void *kind) {
    dhead ("StateChangeTimer_CallBack", DEBUG_CCAPI20_Channel);
    dparams ("%x", kind);
    dassert (pCapiChannel != 0);
    CCAPI20_Channel *pChan = (CCAPI20_Channel *)pCapiChannel;
    switch (pChan->GetState()) {
    case cs_B_ConnectPending:
    case cs_D_Connected:
        if ((tULong)kind == TIMER_FOR_RINGING) {
            dwarning (0); // this kind of timer should be cancelled before
            break;
        }
        // no break!
    case cs_D_ConnectPending:
    case cs_D_ConnectRequest:
        pChan->DisconnectReq();
        break;
    case cs_D_DisconnectPending:
        dwarning (0);
        pChan->Release();
        break;
    default:
        // in this states the timer should have been canceled before
        dwarning (0);
        break;
    }
}


/*===========================================================================*\
    CCAPI20_Channel Konstruktor
\*===========================================================================*/

CCAPI20_Channel::CCAPI20_Channel (void)
  : hStateChangeTimer (vIllegalHandle),
    curState (cs_Released),
    NCPIout (0),
    m_OutstandB3Inds (0),
    RingingTimeout (0) {

    dhead ("CCAPI20_Channel-Constructor", DEBUG_CCAPI20_Channel);
    Protect_Create (&hProtect);
    SetController (IllegalController);
    SetPLCI (IllegalPLCI);
    SetNCCI (IllegalNCCI);
    SetInitiator (vFalse);
    SecTimer_Create (&hStateChangeTimer);
}


/*===========================================================================*\
    CCAPI20_Channel Destruktor
\*===========================================================================*/

CCAPI20_Channel::~CCAPI20_Channel (void) {
    dhead ("CCAPI20_Channel-Destructor", DEBUG_CCAPI20_Channel);
    dassert (GetState() == cs_Released);
    dassert (GetPLCI() == IllegalPLCI);
    dassert (GetNCCI() == IllegalNCCI);
    dwarning (GetInitiator() == vFalse);
    SetNCPI (0);
    if (hStateChangeTimer != vIllegalHandle) {
        SecTimer_Destroy (&hStateChangeTimer);
    }
    Protect_Destroy (&hProtect);
}


/*===========================================================================*\
    CCAPI20_Channel::SetNCPI
\*===========================================================================*/

void CCAPI20_Channel::SetNCPI (c_struct NCPI) {
    dhead ("CCAPI20_Channel::SetNCPI", DEBUG_CCAPI20_Channel);
    Conv_DuplicateStruct (&NCPIout, NCPI);
}


/*===========================================================================*\
    CCAPI20_Channel::AcceptIncomingCall
\*===========================================================================*/

tBool CCAPI20_Channel::AcceptIncomingCall (void) {
    dhead ("CCAPI20_Channel::AcceptIncomingCall", DEBUG_CCAPI20_Channel);
    // TODO: test if AcceptIncommingCall is not called twice and only in the real right state (e.g. before AlertReq)
    dwarning (GetState() == cs_D_ConnectIndication);
    if (GetState() != cs_D_ConnectIndication) {
        RETURN ('x', vFalse);
    }
    dassert (GetPLCI() != IllegalPLCI);
    if (SetNewPLCIEntry (GetPLCI()) == vTrue) {
        RETURN ('x', vTrue);
    }
    DisconnectReq();
    //if (SetState (cs_D_DisconnectPending, cs_D_ConnectIndication) == vTrue) {
    //    dassert (hStateChangeTimer != vIllegalHandle);
    //    SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
    //                   this, (void *)TIMER_FOR_DISCONNECT);
    //    CONNECT_RESP (GetNextMsgNr(), GetPLCI(), &cpV_ConnectResp_Ignore);
    //}
    RETURN ('x', vFalse);
}


/*===========================================================================*\
    CCAPI20_Channel::Register
\*===========================================================================*/

void CCAPI20_Channel::Register (tUInt MaxDataSize) {
    dhead ("CCAPI20_Channel::Register", DEBUG_CCAPI20_Channel);
    dparams ("t=%x", this);
    dassert (GetState() == cs_Released);
    dassert (MaxDataSize <= CAPI_MAX_DATA_SIZE);
    dassert (m_OutstandB3Inds == 0);
    dwarning (MaxDataSize > 100);
    if (  (hStateChangeTimer == vIllegalHandle)
       && (SecTimer_Create (&hStateChangeTimer) == vFalse)) {
        Register_Complete (ci_Int_TimerInitFailure);
        return;
    }
    if (SetState (cs_Registering, cs_Released) == vTrue) {
        REGISTER (1, CAPI_MAX_OUTSTANDING_FRAMES, MaxDataSize);
    } else {
        dwarning (0);
    }
}


/*===========================================================================*\
    CCAPI20_Channel::Release
\*===========================================================================*/

void CCAPI20_Channel::Release (void) {
    dhead ("CCAPI20_Channel::Release", DEBUG_CCAPI20_Channel);
    dassert (GetState() > cs_Released);
    dassert (m_OutstandB3Inds == 0);
    dassert (GetState() < cs_D_DisconnectPending);
    Protect_BeginWrite (hProtect);
    switch (GetState()) {
    case cs_Released:
        dwarning (0);
        Protect_EndWrite (hProtect);
        Release_Complete();
        break;
    case cs_Releasing:
        // nothing to do!
        Protect_EndWrite (hProtect);
        break;
    default:
        dassert (0);
        // no break;
    case cs_Registered:
        curState = cs_Releasing;
        Protect_EndWrite (hProtect);
        ResetValues();
        RELEASE();
    }
}


/*===========================================================================*\
    CCAPI20_Channel::ConnectReq
\*===========================================================================*/

tBool CCAPI20_Channel::ConnectReq (c_dword Controller, cp_connectreq *Params) {
    dhead ("CCAPI20_Channel::ConnectReq", DEBUG_CCAPI20_Channel);
    dassert (Params != 0);
    dassert (m_OutstandB3Inds == 0);
    if (SetState (cs_D_ConnectRequest, cs_Registered) == vFalse) {
        RETURN ('x', vFalse);
    }
    dassert (GetController() == IllegalController);
    SetController (Controller);
    SetInitiator (vTrue);
    CONNECT_REQ (GetNextMsgNr(), Controller, Params);
    RETURN ('x', vTrue);
}


/*===========================================================================*\
    CCAPI20_Channel::DataB3Req
\*===========================================================================*/

void CCAPI20_Channel::DataB3Req (cp_datab3 *Params) {
    dhead ("CCAPI20_Channel::DataB3Req", DEBUG_CCAPI20_Channel);
    dassert (Params != 0);
    dwarning (GetState() == cs_Connected);
    if (GetState() == cs_Connected) {
        dassert (GetNCCI() != IllegalNCCI);
        DATA_B3_REQ (GetNextMsgNr(), GetNCCI(), Params);
    } else {
        cp_datab3conf cp = {Params->DataHandle, ci_Int_NotConnected};
        DataB3Conf (&cp);
    }
}


/*===========================================================================*\
    CCAPI20_Channel::DisconnectReq
\*===========================================================================*/

void CCAPI20_Channel::DisconnectReq (tBool FastDisconnect) {
    dhead ("CCAPI20_Channel::DisconnectReq", DEBUG_CCAPI20_Channel);
    dwarning (GetState() >= cs_Registered);
    Protect_BeginWrite (hProtect);
    switch (GetState()) {
    case cs_D_ConnectRequest:
        curState = cs_D_DisconnectPending;
        dassert (hStateChangeTimer != vIllegalHandle);
        Protect_EndWrite (hProtect);
        SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                       this, (void *)TIMER_FOR_DISCONNECT);
        break;

    case cs_D_ConnectIndication:
        curState = cs_D_DisconnectPending;
        Protect_EndWrite (hProtect);
        SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                       this, (void *)TIMER_FOR_DISCONNECT);
        CONNECT_RESP (GetNextMsgNr(), GetPLCI(), &cpV_ConnectResp_Ignore);
        break;

    case cs_D_DisconnectRespPending:
    case cs_D_DisconnectPending:
        /*----- Do Nothing!! -----*/
        Protect_EndWrite (hProtect);
        break;

    case cs_B_DisconnectRespPending:
    case cs_B_DisconnectPending:
        if (FastDisconnect == vFalse) {
            Protect_EndWrite (hProtect);
            break;
        }
        // no break; -> Okay we make a second test for FastDisconnect before calling DISCONNECT_REQ,
        // but that doesn't matter

    case cs_Connected:
        if (FastDisconnect == vFalse) {
            dassert (GetNCCI() != IllegalNCCI);
            curState = cs_B_DisconnectPending;
            Protect_EndWrite (hProtect);
            DISCONNECT_B3_REQ (GetNextMsgNr(), GetNCCI(), NCPIout);
            break;
        }
        // no break!

    default:
        dassert (GetPLCI() != IllegalPLCI);
        dwarning (GetNCCI() == IllegalNCCI);
        curState = cs_D_DisconnectPending;
        dassert (hStateChangeTimer != vIllegalHandle);
        Protect_EndWrite (hProtect);
        SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                       this, (void *)TIMER_FOR_DISCONNECT);
        DISCONNECT_REQ (GetNextMsgNr(), GetPLCI());
        break;

    case cs_Released:
    case cs_Releasing:
    case cs_Registering:
        dwarning (0); // DisconnectReq should not been called in this states!
        // no break!
    case cs_Registered:
        Protect_EndWrite (hProtect);
        DisconnectInd (ci_NoError);
        break;

    }
}


/*===========================================================================*\
    CCAPI20_Channel::ConnectResp
\*===========================================================================*/

void CCAPI20_Channel::ConnectResp (cp_connectresp *Params) {
    dhead ("CCAPI20_Channel::ConnectResp", DEBUG_CCAPI20_Channel);
    dassert (Params != 0);
    dwarning (GetState() == cs_D_ConnectIndication);
    dassert (GetInitiator() == vFalse);
    dassert (GetPLCI() != IllegalPLCI);
    dassert (m_OutstandB3Inds == 0);
    if (Params->Reject == creject_Accept) {
        if (SetState (cs_D_ConnectPending, cs_D_ConnectIndication) == vTrue) {
            CONNECT_RESP (GetNextMsgNr(), GetPLCI(), Params);
        } else {
            dwarning (0);
        }
    } else {
        if (SetState (cs_D_DisconnectPending, cs_D_ConnectIndication) == vTrue) {
            dassert (hStateChangeTimer != vIllegalHandle);
            SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                           this, (void *)TIMER_FOR_DISCONNECT);
            CONNECT_RESP (GetNextMsgNr(), GetPLCI(), Params);
        } else {
            dwarning (0);
        }
    }
}


/*===========================================================================*\
    CCAPI20_Channel::DataB3Resp
\*===========================================================================*/

void CCAPI20_Channel::DataB3Resp (c_word DataHandle) {
    dhead ("CCAPI20_Channel::DataB3Resp", DEBUG_CCAPI20_Channel);
    dassert (GetNCCI() != IllegalNCCI);
    dwarning (GetState() == cs_Connected);
    dassert (atomicGet (&m_OutstandB3Inds) > 0);

    if (GetState() == cs_Connected) {
        // default way
        atomicDec (&m_OutstandB3Inds);
        DATA_B3_RESP (GetNextMsgNr(), GetNCCI(), DataHandle);

    } else if (atomicDec (&m_OutstandB3Inds) == vFalse) {
        // DisconnectB3Ind received, but some DataB3Resp are still missing
        DATA_B3_RESP (GetNextMsgNr(), GetNCCI(), DataHandle);

    } else {
        // DisconnectB3Ind received and all outstanding DataB3Ind got a responses
        DATA_B3_RESP (GetNextMsgNr(), GetNCCI(), DataHandle);
        cp_disconnectb3ind dummyParams = { ci_NoError, 0 };
        DisconnectB3Resp (&dummyParams);
    }
}


/*===========================================================================*\
    CCAPI20_Channel::REGISTER_COMPLETE
\*===========================================================================*/

void CCAPI20_Channel::REGISTER_COMPLETE (c_info ErrorCode) {
    dhead ("CCAPI20_Channel::REGISTER_COMPLETE", DEBUG_CCAPI20_Channel);
    dwarning (ErrorCode == ci_NoError);
    dassert (GetState() == cs_Registering);
    if (SetState (CapiInfoIsError (ErrorCode) ? cs_Released : cs_Registered, cs_Registering) == vTrue) {
        Register_Complete (ErrorCode);
    } else {
        dwarning (0);
    }
}


/*===========================================================================*\
    CCAPI20_Channel::RELEASE_COMPLETE
\*===========================================================================*/

void CCAPI20_Channel::RELEASE_COMPLETE (void) {
    dhead ("CCAPI20_Channel::RELEASE_COMPLETE", DEBUG_CCAPI20_Channel);
    ResetValues();
    dassert (hStateChangeTimer != vIllegalHandle);
    dinfo ((GetState() == cs_Registered) || (GetState() == cs_Releasing), 
           "ReleaseComplete, but State=%x ", GetState());
    SecTimer_Cancel (hStateChangeTimer);

    Protect_BeginWrite (hProtect);
    tCapiState oldState = curState;
    curState            = cs_Released;
    Protect_EndWrite (hProtect);
    switch (oldState) {
    case cs_Released:
        dassert (0);
        break;
    case cs_Registered:
        break;
    default:
        DisconnectInd (ci_NoError);
        break;
    }
    Release_Complete();
}


/*===========================================================================*\
    CCAPI20_Channel - Conf-Messages
\*===========================================================================*/

void CCAPI20_Channel::CONNECT_CONF (c_word, c_dword PLCI, c_info Info) {
    dhead ("CCAPI20_Channel::CONNECT_CONF", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (CapiInfoIsSuccess (Info)) {
        if (SetState (cs_D_ConnectPending, cs_D_ConnectRequest) == vTrue) {
            dassert (GetController() != IllegalController);
            dassert (GetPLCI() == IllegalPLCI);
            dassert (GetNCCI() == IllegalNCCI);
            dassert (GetInitiator() == vTrue);
            SetPLCI (PLCI);
            if (RingingTimeout) {
                dassert (hStateChangeTimer != vIllegalHandle);
                SecTimer_Call (hStateChangeTimer, RingingTimeout, StateChangeTimer_CallBack,
                               this, (void *)TIMER_FOR_RINGING);
            }
            ConnectConf();
        } else {
            dwarning (0);
            if (GetState() == cs_D_DisconnectPending) {
                dassert (GetPLCI() == IllegalPLCI);
                dassert (GetController() != IllegalController);
                dassert (hStateChangeTimer != vIllegalHandle);
                SetPLCI (PLCI);
                SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                               this, (void *)TIMER_FOR_DISCONNECT);
            }
            DISCONNECT_REQ (GetNextMsgNr(), PLCI);
        }
    } else {
        SetController (IllegalController);
        SetInitiator (vFalse);
        if (SetState (cs_Registered, cs_D_ConnectRequest) == vFalse) {
            if (SetState (cs_Registered, cs_D_DisconnectPending) == vFalse) {
                dassert (0);
            }
        }
        DisconnectInd (Info);
    }
}

void CCAPI20_Channel::CONNECT_B3_CONF (c_word, c_dword NCCI, c_info Info) {
    dhead ("CCAPI20_Channel::CONNECT_B3_CONF", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (CapiInfoIsSuccess (Info)) {
        if ((GetPLCIFromNCCI (NCCI) == GetPLCI()) && (SetState (cs_B_ConnectPending, cs_D_Connected) == vTrue)) {
            dassert (GetNCCI() == IllegalNCCI);
            dassert (GetInitiator() == vFalse);
            SetNCCI (NCCI);
            dprint ("curstate=%x, ncci=%x, NCCI=%x ", GetState(), GetNCCI(), NCCI);
            return;
        }
        dwarning (0);
    }
    DISCONNECT_REQ (GetNextMsgNr(), GetPLCIFromNCCI (NCCI));
}

void CCAPI20_Channel::DATA_B3_CONF (c_word, c_dword NCCI, cp_datab3conf *Params) {
    dhead ("CCAPI20_Channel::DATA_B3_CONF", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    dassert (Params != 0);
    dprint ("Handle=%x Info=%x ", Params->DataHandle, Params->Info);
    if (NCCI != GetNCCI()) {
        dwarning (0);
        DISCONNECT_B3_REQ (GetNextMsgNr(), NCCI, 0);
        return;
    }
    dwarning (GetState() == cs_Connected);
    DataB3Conf (Params);
}

void CCAPI20_Channel::DISCONNECT_B3_CONF (c_word, c_dword NCCI, c_info Info) {
    dhead ("CCAPI20_Channel::DISCONNECT_B3_CONF", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (NCCI == GetNCCI()) {
        dassert (GetState() == cs_B_DisconnectPending);
        if (CapiInfoIsSuccess (Info)) {
            return;
        }
        dwarning (0);
    } else {
        dwarning (0);
    }
    DISCONNECT_REQ (GetNextMsgNr(), GetPLCIFromNCCI (NCCI));
}

void CCAPI20_Channel::DISCONNECT_CONF (c_word, c_dword PLCI, c_info Info) {
    dhead ("CCAPI20_Channel::DISCONNECT_CONF", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    dassert (m_OutstandB3Inds == 0);
    if (PLCI == GetPLCI()) {
        if (CapiInfoIsSuccess (Info)) {
            dassert (GetState() == cs_D_DisconnectPending);
        } else {
            dprintErr ("Info=%x ", Info);
            dassert (hStateChangeTimer != vIllegalHandle);
            SecTimer_Cancel (hStateChangeTimer);
            Release();
        }
    }
}

void CCAPI20_Channel::LISTEN_CONF (c_word, c_dword Controller, c_info Info) {
    dhead ("CCAPI20_Channel::LISTEN_CONF", DEBUG_CCAPI20_Channel);
    ListenConf (Controller, Info);
}

void CCAPI20_Channel::SELECT_B_PROTOCOL_CONF (c_word, c_dword PLCI, c_info Info) {
    dhead ("CCAPI20_Channel::SELECT_B_PROTOCOL_CONF", DEBUG_CCAPI20_Channel);
    dwarning (PLCI == GetPLCI());
    if (PLCI == GetPLCI() && (SetState (cs_ProtocolPending, cs_D_Connected) == vTrue)) {
        SelectBProtocolConf (Info);
        if (GetState() == cs_D_Connected) {
            ConnectB3Req();
        }
    }
}

void CCAPI20_Channel::MANUFACTURER_CONF (c_word, c_dword Controller, cp_manufacturer *Params) {
    dhead ("CCAPI20_Channel::MANUFACTURER_CONF", DEBUG_CCAPI20_Channel);
    ManufacturerConf (Controller, Params);
}


/*===========================================================================*\
    CCAPI20_Channel - Ind-Messages
\*===========================================================================*/

tBool CCAPI20_Channel::CONNECT_IND (c_word, c_dword PLCI, cp_connectind *Params) {
    dhead ("CCAPI20_Channel::CONNECT_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (SetState (cs_D_ConnectIndication, cs_Registered) == vFalse) {
        RETURN ('x', vFalse);
    }
    dassert (GetController() == IllegalController);
    dassert (GetPLCI() == IllegalPLCI);
    dassert (GetNCCI() == IllegalNCCI);
    SetPLCI (PLCI);
    SetController (GetControllerFromPLCI (PLCI));
    ConnectInd (GetController(), Params);
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::CONNECT_ACTIVE_IND (c_word, c_dword PLCI, c_struct ConnectedNumber) {
    dhead ("CCAPI20_Channel::CONNECT_ACTIVE_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if ((PLCI != GetPLCI()) || (SetState (cs_D_Connected, cs_D_ConnectPending) == vFalse)) {
        RETURN ('x', vFalse);
    }
    dassert (hStateChangeTimer != vIllegalHandle);
    CONNECT_ACTIVE_RESP (GetNextMsgNr(), PLCI);
    dassert (hStateChangeTimer != vIllegalHandle);
    SecTimer_Call (hStateChangeTimer, TIMER_BCHANNEL_WAITTIME, StateChangeTimer_CallBack,
                   this, (void *)TIMER_FOR_BCHANNEL);
    ConnectActiveInd (ConnectedNumber);
    if (GetState() != cs_D_Connected) {
        SetInitiator (vFalse);
    } else if (GetInitiator() == vTrue) {
        SetInitiator (vFalse);
        ConnectB3Req();
    }
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::CONNECT_B3_IND (c_word, c_dword NCCI, c_struct) {
    dhead ("CCAPI20_Channel::CONNECT_B3_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if ((GetPLCIFromNCCI (NCCI) != GetPLCI()) || (SetState (cs_B_ConnectPending, cs_D_Connected) == vFalse)) {
        RETURN ('x', vFalse);
    }
    dassert (GetNCCI() == IllegalNCCI);
    dassert (GetInitiator() == vFalse);
    SetNCCI (NCCI);
    cp_connectb3resp ConnB3Params = { creject_Accept, NCPIout };
    CONNECT_B3_RESP (GetNextMsgNr(), GetNCCI(), &ConnB3Params);
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::CONNECT_B3_ACTIVE_IND (c_word, c_dword NCCI, c_struct NCPI) {
    dhead ("CCAPI20_Channel::CONNECT_B3_ACTIVE_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if ((NCCI != GetNCCI()) || (SetState (cs_Connected, cs_B_ConnectPending) == vFalse)) {
        RETURN ('x', vFalse);
    }
    dassert (GetPLCI() != IllegalPLCI);
    dwarning (GetInitiator() == vFalse);
    CONNECT_B3_ACTIVE_RESP (GetNextMsgNr(), NCCI);
    dassert (hStateChangeTimer != vIllegalHandle);
    SecTimer_Cancel (hStateChangeTimer);
    ConnectB3ActiveInd (NCPI);
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::DATA_B3_IND (c_word, c_dword NCCI, cp_datab3 *Params) {
    dhead ("CCAPI20_Channel::DATA_B3_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (NCCI != GetNCCI()) {
        RETURN ('x', vFalse);
    }
    dwarning (GetState() == cs_Connected);
    atomicInc (&m_OutstandB3Inds);
    DataB3Ind (Params);
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::DISCONNECT_B3_IND (c_word, c_dword NCCI, cp_disconnectb3ind *Params) {
    dhead ("CCAPI20_Channel::DISCONNECT_B3_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (NCCI != GetNCCI()) {
        RETURN ('x', vFalse);
    }
    dassert ((GetState() == cs_B_ConnectPending) || (GetState() == cs_Connected) ||
             (GetState() == cs_B_DisconnectPending));
    dassert (GetPLCI() == GetPLCIFromNCCI (NCCI));

    Protect_BeginWrite (hProtect);
    curState = cs_B_DisconnectRespPending;
    Protect_EndWrite (hProtect);
    // TODO: remember Params for a later call to DisconnectB3Ind
    if (!m_OutstandB3Inds) {
        DisconnectB3Resp (Params);
    }
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::DISCONNECT_IND (c_word, c_dword PLCI, c_info Reason) {
    dhead ("CCAPI20_Channel::DISCONNECT_IND", DEBUG_CCAPI20_Channel);
    dparams ("s=%x", GetState());
    if (PLCI != GetPLCI()) {
        RETURN ('x', vFalse);
    }
    dwarning (m_OutstandB3Inds == 0);
    dwarning (GetNCCI() == IllegalNCCI);
    dassert (GetState() > cs_Registered);
    dassert (hStateChangeTimer != vIllegalHandle);
    SecTimer_Cancel (hStateChangeTimer);
    if (SetState (cs_D_DisconnectRespPending, cs_B_DisconnectRespPending) == vFalse) {
        dassert (m_OutstandB3Inds == 0);
        DISCONNECT_RESP (GetNextMsgNr(), PLCI);
        ResetValues();
        Protect_BeginWrite (hProtect);
        curState = cs_Registered;
        Protect_EndWrite (hProtect);
        DisconnectInd (Reason);
    }
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::INFO_IND (c_word, c_dword Address, cp_infoind *Params) {
    dhead ("CCAPI20_Channel::INFO_IND", DEBUG_CCAPI20_Channel);
    dassert (Params != 0);
    // TODO: Handle InfoInd for Controller-Addresses
    if (Address != GetPLCI()) {
        RETURN ('x', vFalse);
    }
    INFO_RESP (GetNextMsgNr(), GetControllerFromPLCI (Address));
    InfoInd (Params);
    RETURN ('x', vTrue);
}

tBool CCAPI20_Channel::MANUFACTURER_IND (c_word, c_dword Controller, cp_manufacturer *Params) {
    dhead ("CCAPI20_Channel::MANUFACTURER_IND", DEBUG_CCAPI20_Channel);
    MANUFACTURER_RESP (GetNextMsgNr(), Controller, Params);
    ManufacturerInd (Controller, Params);
    return vTrue;
}


/*===========================================================================*\
\*===========================================================================*/

inline void CCAPI20_Channel::ResetValues (void) {
    dhead ("CCAPI20_Channel::ResetValues", DEBUG_CCAPI20_Channel);
    Protect_BeginWrite (hProtect);
    FreePLCIEntry (GetPLCI());
    SetPLCI (IllegalPLCI);
    SetNCCI (IllegalNCCI);
    SetController (IllegalController);
    SetInitiator (vFalse);
    Protect_EndWrite (hProtect);
    dwarning (m_OutstandB3Inds == 0);
    atomicSet (&m_OutstandB3Inds, 0);
}


void CCAPI20_Channel::DisconnectB3Resp (cp_disconnectb3ind *Params) {
    dhead ("CCAPI20_Channel::DisconnectB3Resp", DEBUG_CCAPI20_Channel);
    dassert (Params != 0);
    dwarning (GetState() == cs_B_DisconnectRespPending);
    if (SetState (cs_D_Connected, cs_B_DisconnectRespPending) == vTrue) {
        dassert (GetNCCI() != IllegalNCCI);
        DISCONNECT_B3_RESP (GetNextMsgNr(), GetNCCI());
        SetNCCI (IllegalNCCI);
        DisconnectB3Ind (Params);
        if (SetState (cs_D_DisconnectPending, cs_D_Connected) == vTrue) {
            dassert (hStateChangeTimer != vIllegalHandle);
            SecTimer_Call (hStateChangeTimer, TIMER_DISCONNECT_WAITTIME, StateChangeTimer_CallBack,
                           this, (void *)TIMER_FOR_DISCONNECT);
            DISCONNECT_REQ (GetNextMsgNr(), GetPLCI());
        }
    } else if (SetState (cs_Registered, cs_D_DisconnectRespPending) == vTrue) {
        dwarning (0);
        dassert (m_OutstandB3Inds == 0);
        SecTimer_Cancel (hStateChangeTimer);
        DISCONNECT_B3_RESP (GetNextMsgNr(), GetNCCI());
        SetNCCI (IllegalNCCI);
        DISCONNECT_RESP (GetNextMsgNr(), GetPLCI());
        ResetValues();
        DisconnectInd (ci_NoError);
    } else {
        dwarning (0);
    }
}

/*===========================================================================*\
\*===========================================================================*/
